import sempy.fabric as fabric
import numpy as np
from typing import Optional
import sempy_labs._icons as icons


def check_fallback_reason(dataset: str, workspace: Optional[str] = None):
    """
    Shows the reason a table in a Direct Lake semantic model would fallback to DirectQuery.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        The tables in the semantic model and their fallback reason.
    """

    workspace = fabric.resolve_workspace_name(workspace)

    dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    dfP_filt = dfP[dfP["Mode"] == "DirectLake"]

    if len(dfP_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{dataset}' semantic model is not in Direct Lake. This function is only applicable to Direct Lake semantic models."
        )
    else:
        df = fabric.evaluate_dax(
            dataset=dataset,
            workspace=workspace,
            dax_string="""
        SELECT [TableName] AS [Table Name],[FallbackReason] AS [FallbackReasonID]
        FROM $SYSTEM.TMSCHEMA_DELTA_TABLE_METADATA_STORAGES
        """,
        )

        value_mapping = {
            0: "No reason for fallback",
            1: "This table is not framed",
            2: "This object is a view in the lakehouse",
            3: "The table does not exist in the lakehouse",
            4: "Transient error",
            5: "Using OLS will result in fallback to DQ",
            6: "Using RLS will result in fallback to DQ",
        }

        # Create a new column based on the mapping
        df["Fallback Reason Detail"] = np.vectorize(value_mapping.get)(
            df["FallbackReasonID"]
        )

        return df



import sempy.fabric as fabric
import pandas as pd
import re
import datetime
import time
from pyspark.sql import SparkSession
from sempy_labs.tom import connect_semantic_model
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def refresh_calc_tables(dataset: str, workspace: Optional[str] = None):
    """
    Recreates the delta tables in the lakehouse based on the DAX expressions stored as model annotations in the Direct Lake semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    spark = SparkSession.builder.getOrCreate()

    start_time = datetime.datetime.now()
    timeout = datetime.timedelta(minutes=1)
    success = False

    workspace = fabric.resolve_workspace_name(workspace)

    while not success:
        try:
            with connect_semantic_model(
                dataset=dataset, readonly=True, workspace=workspace
            ) as tom:
                success = True
                for a in tom.model.Annotations:
                    if any(a.Name == t.Name for t in tom.model.Tables):
                        tName = a.Name
                        query = a.Value

                        if not query.startswith("EVALUATE"):
                            daxquery = "EVALUATE \n" + query
                        else:
                            daxquery = query

                        try:
                            df = fabric.evaluate_dax(
                                dataset=dataset,
                                dax_string=daxquery,
                                workspace=workspace,
                            )

                            # Update column names for non-field parameters
                            if query.find("NAMEOF") == -1:
                                for old_column_name in df.columns:
                                    pattern = r"\[([^\]]+)\]"

                                    matches = re.findall(pattern, old_column_name)
                                    new_column_name = matches[0]
                                    new_column_name = new_column_name.replace(" ", "")

                                    df.rename(
                                        columns={old_column_name: new_column_name},
                                        inplace=True,
                                    )

                                    # Update data types for lakehouse columns
                                    dataType = next(
                                        str(c.DataType)
                                        for c in tom.all_columns()
                                        if c.Parent.Name == tName
                                        and c.SourceColumn == new_column_name
                                    )
                                    # dfC_type = dfC[(dfC['Table Name'] == tName) & (dfC['Source'] == new_column_name)]
                                    # dataType = dfC_type['Data Type'].iloc[0]

                                    if dataType == "Int64":
                                        df[new_column_name] = df[
                                            new_column_name
                                        ].astype(int)
                                    elif dataType in ["Decimal", "Double"]:
                                        df[new_column_name] = df[
                                            new_column_name
                                        ].astype(float)
                                    elif dataType == "Boolean":
                                        df[new_column_name] = df[
                                            new_column_name
                                        ].astype(bool)
                                    elif dataType == "DateTime":
                                        df[new_column_name] = pd.to_datetime(
                                            df[new_column_name]
                                        )
                                    else:
                                        df[new_column_name] = df[
                                            new_column_name
                                        ].astype(str)
                            # else:
                            #    second_column_name = df.columns[1]
                            #    third_column_name = df.columns[2]
                            #    df[third_column_name] = df[third_column_name].astype(int)

                            # Remove calc columns from field parameters
                            #    mask = df[second_column_name].isin(dfC_filt['Full Column Name'])
                            #    df = df[~mask]

                            delta_table_name = tName.replace(" ", "_")
                            print(
                                f"{icons.in_progress} Refresh of the '{delta_table_name}' table within the lakehouse is in progress..."
                            )

                            spark_df = spark.createDataFrame(df)
                            spark_df.write.mode("overwrite").format(
                                "delta"
                            ).saveAsTable(delta_table_name)
                            print(
                                f"{icons.green_dot} Calculated table '{tName}' has been refreshed as the '{delta_table_name.lower()}' table in the lakehouse."
                            )
                        except Exception as e:
                            raise ValueError(
                                f"{icons.red_dot} Failed to create calculated table '{tName}' as a delta table in the lakehouse."
                            ) from e

        except Exception:
            if datetime.datetime.now() - start_time > timeout:
                break
            time.sleep(1)



from sempy_labs.report._generate_report import (
    create_report_from_reportjson,
    # update_report_from_reportjson,
)
from sempy_labs.report._report_functions import (
    get_report_json,
    # report_dependency_tree,
    export_report,
    clone_report,
    launch_report,
    # list_report_pages,
    # list_report_visuals,
    # list_report_bookmarks,
    # translate_report_titles
)
from sempy_labs.report._report_rebind import (
    report_rebind,
    report_rebind_all,
)

__all__ = [
    "create_report_from_reportjson",
    # "update_report_from_reportjson",
    "get_report_json",
    # report_dependency_tree,
    "export_report",
    "clone_report",
    "launch_report",
    # list_report_pages,
    # list_report_visuals,
    # list_report_bookmarks,
    # translate_report_titles,
    "report_rebind",
    "report_rebind_all",
]



import sempy.fabric as fabric
import json
import os
import shutil
import xml.etree.ElementTree as ET
from sempy_labs._list_functions import list_tables
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
from sempy._utils._log import log
from typing import Optional
import sempy_labs._icons as icons


@log
def create_pqt_file(
    dataset: str,
    workspace: Optional[str] = None,
    file_name: Optional[str] = "PowerQueryTemplate",
):
    """
    Dynamically generates a `Power Query Template <https://learn.microsoft.com/power-query/power-query-template>`_ file based on the semantic model. The .pqt file is
     saved within the Files section of your lakehouse.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    file_name : str, default='PowerQueryTemplate'
        The name of the Power Query Template file to be generated.
    """

    lakeAttach = lakehouse_attached()

    if lakeAttach is False:
        raise ValueError(
            f"{icons.red_dot} In order to run the 'create_pqt_file' function, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
        )

    workspace = fabric.resolve_workspace_name(workspace)

    folderPath = "/lakehouse/default/Files"
    subFolderPath = os.path.join(folderPath, "pqtnewfolder")
    os.makedirs(subFolderPath, exist_ok=True)

    dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    dfT = list_tables(dataset, workspace)
    dfE = fabric.list_expressions(dataset=dataset, workspace=workspace)

    # Check if M-partitions are used
    if any(dfP["Source Type"] == "M"):

        class QueryMetadata:
            def __init__(
                self,
                QueryName,
                QueryGroupId=None,
                LastKnownIsParameter=None,
                LastKnownResultTypeName=None,
                LoadEnabled=True,
                IsHidden=False,
            ):
                self.QueryName = QueryName
                self.QueryGroupId = QueryGroupId
                self.LastKnownIsParameter = LastKnownIsParameter
                self.LastKnownResultTypeName = LastKnownResultTypeName
                self.LoadEnabled = LoadEnabled
                self.IsHidden = IsHidden

        class RootObject:
            def __init__(
                self, DocumentLocale, EngineVersion, QueriesMetadata, QueryGroups=None
            ):
                if QueryGroups is None:
                    QueryGroups = []
                self.DocumentLocale = DocumentLocale
                self.EngineVersion = EngineVersion
                self.QueriesMetadata = QueriesMetadata
                self.QueryGroups = QueryGroups

        # STEP 1: Create MashupDocument.pq
        mdfileName = "MashupDocument.pq"
        mdFilePath = os.path.join(subFolderPath, mdfileName)
        sb = "section Section1;"
        for table_name in dfP["Table Name"].unique():
            tName = '#"' + table_name + '"'
            sourceExpression = dfT.loc[
                (dfT["Name"] == table_name), "Source Expression"
            ].iloc[0]
            refreshPolicy = dfT.loc[(dfT["Name"] == table_name), "Refresh Policy"].iloc[
                0
            ]
            sourceType = dfP.loc[(dfP["Table Name"] == table_name), "Source Type"].iloc[
                0
            ]

            if sourceType == "M" or refreshPolicy:
                sb = sb + "\n" + "shared " + tName + " = "

            partitions_in_table = dfP.loc[
                dfP["Table Name"] == table_name, "Partition Name"
            ].unique()

            i = 1
            for partition_name in partitions_in_table:
                pSourceType = dfP.loc[
                    (dfP["Table Name"] == table_name)
                    & (dfP["Partition Name"] == partition_name),
                    "Source Type",
                ].iloc[0]
                pQuery = dfP.loc[
                    (dfP["Table Name"] == table_name)
                    & (dfP["Partition Name"] == partition_name),
                    "Query",
                ].iloc[0]

                if pQuery is not None:
                    pQueryNoSpaces = (
                        pQuery.replace(" ", "")
                        .replace("\n", "")
                        .replace("\t", "")
                        .replace("\r", "")
                    )
                    if pQueryNoSpaces.startswith('letSource=""'):
                        pQuery = 'let\n\tSource = ""\nin\n\tSource'

                if pSourceType == "M" and i == 1:
                    sb = sb + pQuery + ";"
                elif refreshPolicy and i == 1:
                    sb = sb + sourceExpression + ";"
                i += 1

        for index, row in dfE.iterrows():
            expr = row["Expression"]
            eName = row["Name"]
            eName = '#"' + eName + '"'
            sb = sb + "\n" + "shared " + eName + " = " + expr + ";"

        with open(mdFilePath, "w") as file:
            file.write(sb)

        # STEP 2: Create the MashupMetadata.json file
        mmfileName = "MashupMetadata.json"
        mmFilePath = os.path.join(subFolderPath, mmfileName)
        queryMetadata = []

        for tName in dfP["Table Name"].unique():
            sourceType = dfP.loc[(dfP["Table Name"] == tName), "Source Type"].iloc[0]
            refreshPolicy = dfT.loc[(dfT["Name"] == tName), "Refresh Policy"].iloc[0]
            if sourceType == "M" or refreshPolicy:
                queryMetadata.append(
                    QueryMetadata(tName, None, None, None, True, False)
                )

        for i, r in dfE.iterrows():
            eName = r["Name"]
            eKind = r["Kind"]
            if eKind == "M":
                queryMetadata.append(
                    QueryMetadata(eName, None, None, None, True, False)
                )
            else:
                queryMetadata.append(
                    QueryMetadata(eName, None, None, None, False, False)
                )

        rootObject = RootObject("en-US", "2.126.453.0", queryMetadata)

        def obj_to_dict(obj):
            if isinstance(obj, list):
                return [obj_to_dict(e) for e in obj]
            elif hasattr(obj, "__dict__"):
                return {k: obj_to_dict(v) for k, v in obj.__dict__.items()}
            else:
                return obj

        jsonContent = json.dumps(obj_to_dict(rootObject), indent=4)

        with open(mmFilePath, "w") as json_file:
            json_file.write(jsonContent)

        # STEP 3: Create Metadata.json file
        mFileName = "Metadata.json"
        mFilePath = os.path.join(subFolderPath, mFileName)
        metaData = {"Name": "fileName", "Description": "", "Version": "1.0.0.0"}
        jsonContent = json.dumps(metaData, indent=4)

        with open(mFilePath, "w") as json_file:
            json_file.write(jsonContent)

        # STEP 4: Create [Content_Types].xml file:
        ns = "http://schemas.openxmlformats.org/package/2006/content-types"
        ET.register_namespace("", ns)
        types = ET.Element("{%s}Types" % ns)
        # default1 = ET.SubElement(
        #    types,
        #    "{%s}Default" % ns,
        #    {"Extension": "json", "ContentType": "application/json"},
        # )
        # default2 = ET.SubElement(
        #    types,
        #    "{%s}Default" % ns,
        #    {"Extension": "pq", "ContentType": "application/x-ms-m"},
        # )
        xmlDocument = ET.ElementTree(types)
        xmlFileName = "[Content_Types].xml"
        xmlFilePath = os.path.join(subFolderPath, xmlFileName)
        xmlDocument.write(
            xmlFilePath, xml_declaration=True, encoding="utf-8", method="xml"
        )

        # STEP 5: Zip up the 4 files
        zipFileName = file_name + ".zip"
        zipFilePath = os.path.join(folderPath, zipFileName)
        shutil.make_archive(zipFilePath[:-4], "zip", subFolderPath)

        # STEP 6: Convert the zip file back into a .pqt file
        newExt = ".pqt"
        directory = os.path.dirname(zipFilePath)
        fileNameWithoutExtension = os.path.splitext(os.path.basename(zipFilePath))[0]
        newFilePath = os.path.join(directory, fileNameWithoutExtension + newExt)
        shutil.move(zipFilePath, newFilePath)

        # STEP 7: Delete subFolder directory which is no longer needed
        shutil.rmtree(subFolderPath, ignore_errors=True)

        print(
            f"{icons.green_dot} '{file_name}.pqt' has been created based on the '{dataset}' semantic model in the '{workspace}' workspace within the Files section of your lakehouse."
        )

    else:
        print(
            f"{icons.yellow_dot} The '{dataset}' semantic model in the '{workspace}' workspace does not use Power Query so a Power Query Template file cannot be generated."
        )



import sempy
import sempy.fabric as fabric
import pandas as pd
from synapse.ml.services.openai import OpenAICompletion
from pyspark.sql.functions import col
from pyspark.sql import SparkSession
from typing import List, Optional, Union
from IPython.display import display
import sempy_labs._icons as icons


def optimize_semantic_model(dataset: str, workspace: Optional[str] = None):

    from ._model_bpa import run_model_bpa
    from .directlake._fallback import check_fallback_reason
    from ._helper_functions import format_dax_object_name

    modelBPA = run_model_bpa(
        dataset=dataset, workspace=workspace, return_dataframe=True
    )
    dfC = fabric.list_columns(dataset=dataset, workspace=workspace, extended=True)
    dfC["Column Object"] = format_dax_object_name(dfC["Table Name"], dfC["Column Name"])
    dfC["Total Size"] = dfC["Total Size"].astype("int")
    dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)

    modelBPA_col = modelBPA[modelBPA["Object Type"] == "Column"]
    modelBPA_col = pd.merge(
        modelBPA_col,
        dfC[["Column Object", "Total Size"]],
        left_on="Object Name",
        right_on="Column Object",
        how="left",
    )

    isDirectLake = any(r["Mode"] == "DirectLake" for i, r in dfP.iterrows())

    if isDirectLake:
        fallback = check_fallback_reason(dataset=dataset, workspace=workspace)
        fallback_filt = fallback[fallback["FallbackReasonID"] == 2]

        if len(fallback_filt) > 0:
            print(
                f"{icons.yellow_dot} The '{dataset}' semantic model is a Direct Lake semantic model which contains views. "
                "Since views always fall back to DirectQuery, it is recommended to only use lakehouse tables and not views."
            )

    # Potential model reduction estimate
    ruleNames = [
        "Remove unnecessary columns",
        "Set IsAvailableInMdx to false on non-attribute columns",
    ]

    for rule in ruleNames:
        df = modelBPA_col[modelBPA_col["Rule Name"] == rule]
        df_filt = df[["Object Name", "Total Size"]].sort_values(
            by="Total Size", ascending=False
        )
        totSize = df["Total Size"].sum()
        if len(df_filt) > 0:
            print(
                f"{icons.yellow_dot} Potential savings of {totSize} bytes from following the '{rule}' rule."
            )
            display(df_filt)
        else:
            print(f"{icons.green_dot} The '{rule}' rule has been followed.")


def generate_measure_descriptions(
    dataset: str,
    measures: Union[str, List[str]],
    gpt_model: Optional[str] = "gpt-35-turbo",
    workspace: Optional[str] = None,
):

    service_name = "synapseml-openai"

    if isinstance(measures, str):
        measures = [measures]

    validModels = ["gpt-35-turbo", "gpt-35-turbo-16k", "gpt-4"]
    if gpt_model not in validModels:
        raise ValueError(
            f"{icons.red_dot} The '{gpt_model}' model is not a valid model. Enter a gpt_model from this list: {validModels}."
        )

    dfM = fabric.list_measures(dataset=dataset, workspace=workspace)

    if measures is not None:
        dfM_filt = dfM[dfM["Measure Name"].isin(measures)]
    else:
        dfM_filt = dfM

    df = dfM_filt[["Table Name", "Measure Name", "Measure Expression"]]

    df["prompt"] = (
        "The following is DAX code used by Microsoft Power BI. Please explain this code in simple terms:"
        + df["Measure Expression"]
    )

    # Generate new column in df dataframe which has the AI-generated descriptions
    completion = {
        OpenAICompletion()
        .setDeploymentName(gpt_model)
        .setMaxTokens(200)
        .setCustomServiceName(service_name)
        .setPromptCol("prompt")
        .setErrorCol("error")
        .setOutputCol("completions")
    }

    completed_df = completion.transform(df).cache()
    completed_df.select(
        col("prompt"),
        col("error"),
        col("completions.choices.text").getItem(0).alias("text"),
    )

    # Update the model to use the new descriptions
    # with connect_semantic_model(dataset=dataset, workspace=workspace, readonly=False) as tom:

    # for t in m.Tables:
    # tName = t.Name
    # for ms in t.Measures:
    # mName = ms.Name
    # mDesc = promptValue

    # m.SaveChanges()


def generate_aggs(
    dataset: str,
    table_name: str,
    columns: Union[str, List[str]],
    workspace: Optional[str] = None,
    lakehouse_workspace: Optional[str] = None,
):

    from ._helper_functions import (
        get_direct_lake_sql_endpoint,
        create_abfss_path,
        format_dax_object_name,
        resolve_lakehouse_id,
    )

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM
    import System

    # columns = {
    # 'SalesAmount': 'Sum',
    # 'ProductKey': 'GroupBy',
    # 'OrderDateKey': 'GroupBy'
    #  }

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    if lakehouse_workspace is None:
        lakehouse_workspace = workspace
        lakehouse_workspace_id = workspace_id
    else:
        lakehouse_workspace_id = fabric.resolve_workspace_id(lakehouse_workspace)

    if isinstance(columns, str):
        columns = [columns]

    columnValues = columns.keys()

    aggTypes = ["Sum", "Count", "Min", "Max", "GroupBy"]
    aggTypesAggregate = ["Sum", "Count", "Min", "Max"]
    numericTypes = ["Int64", "Double", "Decimal"]

    if any(value not in aggTypes for value in columns.values()):
        raise ValueError(
            f"{icons.red_dot} Invalid aggregation type(s) have been specified in the 'columns' parameter. Valid aggregation types: {aggTypes}."
        )

    dfC = fabric.list_columns(dataset=dataset, workspace=workspace)
    dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    dfM = fabric.list_measures(dataset=dataset, workspace=workspace)
    dfR = fabric.list_relationships(dataset=dataset, workspace=workspace)
    if not any(r["Mode"] == "DirectLake" for i, r in dfP.iterrows()):
        raise ValueError(
            f"{icons.red_dot} The '{dataset}' semantic model within the '{workspace}' workspace is not in Direct Lake mode. This function is only relevant for Direct Lake semantic models."
        )

    dfC_filtT = dfC[dfC["Table Name"] == table_name]

    if len(dfC_filtT) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{table_name}' table does not exist in the '{dataset}' semantic model within the '{workspace}' workspace."
        )

    dfC_filt = dfC[
        (dfC["Table Name"] == table_name) & (dfC["Column Name"].isin(columnValues))
    ]

    if len(columns) != len(dfC_filt):
        raise ValueError(
            f"{icons.red_dot} Columns listed in '{columnValues}' do not exist in the '{table_name}' table in the '{dataset}' semantic model within the '{workspace}' workspace."
        )

    # Check if doing sum/count/min/max etc. on a non-number column
    for cm, agg in columns.items():
        dfC_col = dfC_filt[dfC_filt["Column Name"] == cm]
        dataType = dfC_col["Data Type"].iloc[0]
        if agg in aggTypesAggregate and dataType not in numericTypes:
            raise ValueError(
                f"{icons.red_dot} The '{cm}' column in the '{table_name}' table is of '{dataType}' data type. Only columns of '{numericTypes}' data types"
                f" can be aggregated as '{aggTypesAggregate}' aggregation types."
            )

    # Create/update lakehouse delta agg table
    aggSuffix = "_agg"
    aggTableName = f"{table_name}{aggSuffix}"
    aggLakeTName = aggTableName.lower().replace(" ", "_")
    dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    dfP_filt = dfP[dfP["Table Name"] == table_name]
    lakeTName = dfP_filt["Query"].iloc[0]

    sqlEndpointId = get_direct_lake_sql_endpoint(dataset=dataset, workspace=workspace)

    dfI = fabric.list_items(workspace=lakehouse_workspace, type="SQLEndpoint")
    dfI_filt = dfI[(dfI["Id"] == sqlEndpointId)]

    if len(dfI_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The lakehouse (SQL Endpoint) used by the '{dataset}' semantic model does not reside in"
            f" the '{lakehouse_workspace}' workspace. Please update the lakehouse_workspace parameter."
        )

    lakehouseName = dfI_filt["Display Name"].iloc[0]
    lakehouse_id = resolve_lakehouse_id(
        lakehouse=lakehouseName, workspace=lakehouse_workspace
    )

    # Generate SQL query
    query = "SELECT"
    groupBy = "\nGROUP BY"
    for cm, agg in columns.items():
        colFilt = dfC_filt[dfC_filt["Column Name"] == cm]
        sourceCol = colFilt["Source"].iloc[0]

        if agg == "GroupBy":
            query = f"{query}\n{sourceCol},"
            groupBy = f"{groupBy}\n{sourceCol},"
        else:
            query = f"{query}\n{agg}({sourceCol}) AS {sourceCol},"

    query = query[:-1]

    spark = SparkSession.builder.getOrCreate()
    fromTablePath = create_abfss_path(
        lakehouse_id=lakehouse_id,
        lakehouse_workspace_id=lakehouse_workspace_id,
        delta_table_name=lakeTName,
    )
    df = spark.read.format("delta").load(fromTablePath)
    tempTableName = "delta_table_" + lakeTName
    df.createOrReplaceTempView(tempTableName)
    sqlQuery = f"{query} \n FROM {tempTableName} {groupBy}"

    sqlQuery = sqlQuery[:-1]
    print(sqlQuery)

    # Save query to spark dataframe
    spark_df = spark.sql(sqlQuery)
    f"\nCreating/updating the '{aggLakeTName}' table in the lakehouse..."
    # Write spark dataframe to delta table
    aggFilePath = create_abfss_path(
        lakehouse_id=lakehouse_id,
        lakehouse_workspace_id=lakehouse_workspace_id,
        delta_table_name=aggLakeTName,
    )
    spark_df.write.mode("overwrite").format("delta").save(aggFilePath)
    f"{icons.green_dot} The '{aggLakeTName}' table has been created/updated in the lakehouse."

    # Create/update semantic model agg table
    tom_server = fabric.create_tom_server(readonly=False, workspace=workspace)
    m = tom_server.Databases.GetByName(dataset).Model
    print(f"\n{icons.in_progress} Updating the '{dataset}' semantic model...")
    dfC_agg = dfC[dfC["Table Name"] == aggTableName]

    if len(dfC_agg) == 0:
        print(f"{icons.in_progress} Creating the '{aggTableName}' table...")
        exp = m.Expressions["DatabaseQuery"]
        tbl = TOM.Table()
        tbl.Name = aggTableName
        tbl.IsHidden = True

        ep = TOM.EntityPartitionSource()
        ep.Name = aggTableName
        ep.EntityName = aggLakeTName
        ep.ExpressionSource = exp

        part = TOM.Partition()
        part.Name = aggTableName
        part.Source = ep
        part.Mode = TOM.ModeType.DirectLake

        tbl.Partitions.Add(part)

        for i, r in dfC_filt.iterrows():
            scName = r["Source"]
            cName = r["Column Name"]
            dType = r["Data Type"]

            col = TOM.DataColumn()
            col.Name = cName
            col.IsHidden = True
            col.SourceColumn = scName
            col.DataType = System.Enum.Parse(TOM.DataType, dType)

            tbl.Columns.Add(col)
            print(
                f"{icons.green_dot} The '{aggTableName}'[{cName}] column has been added to the '{dataset}' semantic model."
            )

        m.Tables.Add(tbl)
        print(
            f"{icons.green_dot} The '{aggTableName}' table has been added to the '{dataset}' semantic model."
        )
    else:
        print(f"{icons.in_progress} Updating the '{aggTableName}' table's columns...")
        # Remove existing columns
        for t in m.Tables:
            tName = t.Name
            for c in t.Columns:
                cName = c.Name
                if t.Name == aggTableName:
                    m.Tables[tName].Columns.Remove(cName)
        # Add columns
        for i, r in dfC_filt.iterrows():
            scName = r["Source"]
            cName = r["Column Name"]
            dType = r["Data Type"]

            col = TOM.DataColumn()
            col.Name = cName
            col.IsHidden = True
            col.SourceColumn = scName
            col.DataType = System.Enum.Parse(TOM.DataType, dType)

            m.Tables[aggTableName].Columns.Add(col)
            print(
                f"{icons.green_dot} The '{aggTableName}'[{cName}] column has been added."
            )

    # Create relationships
    relMap = {"m": "Many", "1": "One", "0": "None"}

    print(f"\n{icons.in_progress} Generating necessary relationships...")
    for i, r in dfR.iterrows():
        fromTable = r["From Table"]
        fromColumn = r["From Column"]
        toTable = r["To Table"]
        toColumn = r["To Column"]
        cfb = r["Cross Filtering Behavior"]
        sfb = r["Security Filtering Behavior"]
        mult = r["Multiplicity"]

        crossFB = System.Enum.Parse(TOM.CrossFilteringBehavior, cfb)
        secFB = System.Enum.Parse(TOM.SecurityFilteringBehavior, sfb)
        fromCardinality = System.Enum.Parse(
            TOM.RelationshipEndCardinality, relMap.get(mult[0])
        )
        toCardinality = System.Enum.Parse(
            TOM.RelationshipEndCardinality, relMap.get(mult[-1])
        )

        rel = TOM.SingleColumnRelationship()
        rel.FromCardinality = fromCardinality
        rel.ToCardinality = toCardinality
        rel.IsActive = r["Active"]
        rel.CrossFilteringBehavior = crossFB
        rel.SecurityFilteringBehavior = secFB
        rel.RelyOnReferentialIntegrity = r["Rely On Referential Integrity"]

        if fromTable == table_name:
            try:
                rel.FromColumn = m.Tables[aggTableName].Columns[fromColumn]
                m.Relationships.Add(rel)
                print(
                    f"{icons.green_dot} '{aggTableName}'[{fromColumn}] -> '{toTable}'[{toColumn}] relationship has been added."
                )
            except Exception as e:
                print(
                    f"{icons.red_dot} '{aggTableName}'[{fromColumn}] -> '{toTable}'[{toColumn}] relationship has not been created."
                )
                print(f"Exception occured: {e}")
        elif toTable == table_name:
            try:
                rel.ToColumn = m.Tables[aggTableName].Columns[toColumn]
                m.Relationships.Add(rel)
                print(
                    f"{icons.green_dot} '{fromTable}'[{fromColumn}] -> '{aggTableName}'[{toColumn}] relationship has been added."
                )
            except Exception as e:
                print(
                    f"{icons.red_dot} '{fromTable}'[{fromColumn}] -> '{aggTableName}'[{toColumn}] relationship has not been created."
                )
                print(f"Exception occured: {e}")
    "Relationship creation is complete."

    # Create IF measure
    f"\n{icons.in_progress} Creating measure to check if the agg table can be used..."
    aggChecker = "IF("
    dfR_filt = dfR[
        (dfR["From Table"] == table_name) & (~dfR["From Column"].isin(columnValues))
    ]

    for i, r in dfR_filt.iterrows():
        toTable = r["To Table"]
        aggChecker = f"{aggChecker}\nISCROSSFILTERED('{toTable}') ||"

    aggChecker = aggChecker[:-3]
    aggChecker = f"{aggChecker},1,0)"
    print(aggChecker)

    # Todo: add IFISFILTERED clause for columns
    f"\n{icons.in_progress} Creating the base measures in the agg table..."
    # Create base agg measures
    dep = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
        SELECT 
         [TABLE] AS [Table Name]
        ,[OBJECT] AS [Object Name]
        ,[OBJECT_TYPE] AS [Object Type]
        ,[REFERENCED_TABLE] AS [Referenced Table]
        ,[REFERENCED_OBJECT] AS [Referenced Object]
        ,[REFERENCED_OBJECT_TYPE] AS [Referenced Object Type]
        FROM $SYSTEM.DISCOVER_CALC_DEPENDENCY
        WHERE [OBJECT_TYPE] = 'MEASURE'
        """,
    )

    baseMeasures = dep[
        (dep["Referenced Object Type"] == "COLUMN")
        & (dep["Referenced Table"] == table_name)
        & (dep["Referenced Object"].isin(columnValues))
    ]
    for i, r in baseMeasures.iterrows():
        tName = r["Table Name"]
        mName = r["Object Name"]
        cName = r["Referenced Object"]
        dfM_filt = dfM[dfM["Measure Name"] == mName]
        expr = dfM_filt["Measure Expression"].iloc[0]

        colFQNonAgg = format_dax_object_name(tName, cName)
        colFQAgg = format_dax_object_name(aggTableName, cName)
        colNQNonAgg = f"{tName}[{cName}]"

        if " " in tName:
            newExpr = expr.replace(colFQNonAgg, colFQAgg)
        else:
            newExpr = expr.replace(colFQNonAgg, colFQAgg).replace(colNQNonAgg, colFQAgg)
        print(expr)
        print(newExpr)

        aggMName = mName + aggSuffix
        measure = TOM.Measure()
        measure.Name = aggMName
        measure.IsHidden = True
        measure.Expression = newExpr
        m.Tables[aggTableName].Measures.Add(measure)
        f"The '{aggMName}' measure has been created in the '{aggTableName}' table."

    # Update base detail measures

    # m.SaveChanges()


# Identify views used within Direct Lake model
# workspace = 'MK Demo 6'
# lakehouse = 'MyLakehouse'
# dataset = 'MigrationTest'
# lakehouse_workspace = workspace

# dfView = pd.DataFrame(columns=['Workspace Name', 'Lakehouse Name', 'View Name'])
# dfP = fabric.list_partitions(dataset = dataset, workspace = workspace)
# isDirectLake = any(r['Mode'] == 'DirectLake' for i, r in dfP.iterrows())

# spark = SparkSession.builder.getOrCreate()
# views = spark.sql(f"SHOW VIEWS IN {lakehouse}").collect()
# for view in views:
#    viewName = view['viewName']
#    isTemporary = view['isTemporary']
#    new_data = {'Workspace Name': workspace, 'Lakehouse Name': lakehouse, 'View Name': viewName}
#    dfView = pd.concat([dfView, pd.DataFrame(new_data, index=[0])], ignore_index=True)
# dfView
# lakeT = get_lakehouse_tables(lakehouse, lakehouse_workspace)
# if not dfP['Query'].isin(lakeT['Table Name'].values):
#    if



import sempy.fabric as fabric
import pandas as pd
import warnings
import datetime
from IPython.display import display, HTML
from pyspark.sql import SparkSession
from sempy_labs._model_dependencies import get_model_calc_dependencies
from sempy_labs._helper_functions import (
    format_dax_object_name,
    resolve_lakehouse_name,
    create_relationship_name,
)
from sempy_labs.lakehouse._get_lakehouse_tables import get_lakehouse_tables
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
from sempy_labs.tom import connect_semantic_model
from sempy_labs._model_bpa_rules import model_bpa_rules
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def run_model_bpa(
    dataset: str,
    rules: Optional[pd.DataFrame] = None,
    workspace: Optional[str] = None,
    export: Optional[bool] = False,
    return_dataframe: Optional[bool] = False,
    extended: Optional[bool] = False,
    **kwargs,
):
    """
    Displays an HTML visualization of the results of the Best Practice Analyzer scan for a semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    rules : pandas.DataFrame, default=None
        A pandas dataframe containing rules to be evaluated.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    export : bool, default=False
        If True, exports the resulting dataframe to a delta table in the lakehouse attached to the notebook.
    return_dataframe : bool, default=False
        If True, returns a pandas dataframe instead of the visualization.
    extended : bool, default=False
        If True, runs the set_vertipaq_annotations function to collect Vertipaq Analyzer statistics to be used in the analysis of the semantic model.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe in HTML format showing semantic model objects which violated the best practice analyzer rules.
    """

    if "extend" in kwargs:
        print(
            "The 'extend' parameter has been deprecated. Please remove this parameter from the function going forward."
        )
        del kwargs["extend"]

    warnings.filterwarnings(
        "ignore",
        message="This pattern is interpreted as a regular expression, and has match groups.",
    )

    workspace = fabric.resolve_workspace_name(workspace)

    if extended:
        with connect_semantic_model(
            dataset=dataset, workspace=workspace, readonly=False
        ) as tom:
            tom.set_vertipaq_annotations()

    with connect_semantic_model(
        dataset=dataset, workspace=workspace, readonly=True
    ) as tom:

        dep = get_model_calc_dependencies(dataset=dataset, workspace=workspace)

        if rules is None:
            rules = model_bpa_rules(
                dataset=dataset, workspace=workspace, dependencies=dep
            )

        rules["Severity"].replace("Warning", "⚠️", inplace=True)
        rules["Severity"].replace("Error", "\u274C", inplace=True)
        rules["Severity"].replace("Info", "ℹ️", inplace=True)

        pd.set_option("display.max_colwidth", 1000)

        violations = pd.DataFrame(columns=["Object Name", "Scope", "Rule Name"])

        scope_to_dataframe = {
            "Relationship": (
                tom.model.Relationships,
                lambda obj: create_relationship_name(
                    obj.FromTable.Name,
                    obj.FromColumn.Name,
                    obj.ToTable.Name,
                    obj.ToColumn.Name,
                ),
            ),
            "Column": (
                tom.all_columns(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Measure": (tom.all_measures(), lambda obj: obj.Name),
            "Hierarchy": (
                tom.all_hierarchies(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Table": (tom.model.Tables, lambda obj: obj.Name),
            "Role": (tom.model.Roles, lambda obj: obj.Name),
            "Model": (tom.model, lambda obj: obj.Model.Name),
            "Calculation Item": (
                tom.all_calculation_items(),
                lambda obj: format_dax_object_name(obj.Parent.Table.Name, obj.Name),
            ),
            "Row Level Security": (
                tom.all_rls(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Partition": (
                tom.all_partitions(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
        }

        for i, r in rules.iterrows():
            ruleName = r["Rule Name"]
            expr = r["Expression"]
            scopes = r["Scope"]

            if isinstance(scopes, str):
                scopes = [scopes]

            for scope in scopes:
                func = scope_to_dataframe[scope][0]
                nm = scope_to_dataframe[scope][1]

                if scope == "Model":
                    x = []
                    if expr(func):
                        x = ["Model"]
                elif scope == "Measure":
                    x = [nm(obj) for obj in tom.all_measures() if expr(obj)]
                elif scope == "Column":
                    x = [nm(obj) for obj in tom.all_columns() if expr(obj)]
                elif scope == "Partition":
                    x = [nm(obj) for obj in tom.all_partitions() if expr(obj)]
                elif scope == "Hierarchy":
                    x = [nm(obj) for obj in tom.all_hierarchies() if expr(obj)]
                elif scope == "Table":
                    x = [nm(obj) for obj in tom.model.Tables if expr(obj)]
                elif scope == "Relationship":
                    x = [nm(obj) for obj in tom.model.Relationships if expr(obj)]
                elif scope == "Role":
                    x = [nm(obj) for obj in tom.model.Roles if expr(obj)]
                elif scope == "Row Level Security":
                    x = [nm(obj) for obj in tom.all_rls() if expr(obj)]
                elif scope == "Calculation Item":
                    x = [nm(obj) for obj in tom.all_calculation_items() if expr(obj)]

                if len(x) > 0:
                    new_data = {"Object Name": x, "Scope": scope, "Rule Name": ruleName}
                    violations = pd.concat(
                        [violations, pd.DataFrame(new_data)], ignore_index=True
                    )

        prepDF = pd.merge(
            violations,
            rules[["Rule Name", "Category", "Severity", "Description", "URL"]],
            left_on="Rule Name",
            right_on="Rule Name",
            how="left",
        )
        prepDF.rename(columns={"Scope": "Object Type"}, inplace=True)
        finalDF = prepDF[
            [
                "Category",
                "Rule Name",
                "Severity",
                "Object Type",
                "Object Name",
                "Description",
                "URL",
            ]
        ]

    if export:
        lakeAttach = lakehouse_attached()
        if lakeAttach is False:
            raise ValueError(
                f"{icons.red_dot} In order to save the Best Practice Analyzer results, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
            )

        dfExport = finalDF.copy()
        delta_table_name = "modelbparesults"

        lakehouse_id = fabric.get_lakehouse_id()
        lakehouse = resolve_lakehouse_name(
            lakehouse_id=lakehouse_id, workspace=workspace
        )

        lakeT = get_lakehouse_tables(lakehouse=lakehouse, workspace=workspace)
        lakeT_filt = lakeT[lakeT["Table Name"] == delta_table_name]

        dfExport["Severity"].replace("⚠️", "Warning", inplace=True)
        dfExport["Severity"].replace("\u274C", "Error", inplace=True)
        dfExport["Severity"].replace("ℹ️", "Info", inplace=True)

        spark = SparkSession.builder.getOrCreate()
        query = f"SELECT MAX(RunId) FROM {lakehouse}.{delta_table_name}"

        if len(lakeT_filt) == 0:
            runId = 1
        else:
            dfSpark = spark.sql(query)
            maxRunId = dfSpark.collect()[0][0]
            runId = maxRunId + 1

        now = datetime.datetime.now()
        dfExport["Workspace Name"] = workspace
        dfExport["Dataset Name"] = dataset
        dfExport["Timestamp"] = now
        dfExport["RunId"] = runId

        dfExport["RunId"] = dfExport["RunId"].astype("int")

        colName = "Workspace Name"
        dfExport.insert(0, colName, dfExport.pop(colName))
        colName = "Dataset Name"
        dfExport.insert(1, colName, dfExport.pop(colName))

        dfExport.columns = dfExport.columns.str.replace(" ", "_")
        spark_df = spark.createDataFrame(dfExport)
        spark_df.write.mode("append").format("delta").saveAsTable(delta_table_name)
        print(
            f"{icons.green_dot} Model Best Practice Analyzer results for the '{dataset}' semantic model have been appended to the '{delta_table_name}' delta table."
        )

    if return_dataframe:
        return finalDF

    pd.set_option("display.max_colwidth", 100)

    finalDF = (
        finalDF[
            [
                "Category",
                "Rule Name",
                "Object Type",
                "Object Name",
                "Severity",
                "Description",
                "URL",
            ]
        ]
        .sort_values(["Category", "Rule Name", "Object Type", "Object Name"])
        .set_index(["Category", "Rule Name"])
    )

    bpa2 = finalDF.reset_index()
    bpa_dict = {
        cat: bpa2[bpa2["Category"] == cat].drop("Category", axis=1)
        for cat in bpa2["Category"].drop_duplicates().values
    }

    styles = """
    <style>
        .tab { overflow: hidden; border: 1px solid #ccc; background-color: #f1f1f1; }
        .tab button { background-color: inherit; float: left; border: none; outline: none; cursor: pointer; padding: 14px 16px; transition: 0.3s; }
        .tab button:hover { background-color: #ddd; }
        .tab button.active { background-color: #ccc; }
        .tabcontent { display: none; padding: 6px 12px; border: 1px solid #ccc; border-top: none; }
        .tabcontent.active { display: block; }
        .tooltip { position: relative; display: inline-block; }
        .tooltip .tooltiptext { visibility: hidden; width: 300px; background-color: #555; color: #fff; text-align: center; border-radius: 6px; padding: 5px; position: absolute; z-index: 1; bottom: 125%; left: 50%; margin-left: -110px; opacity: 0; transition: opacity 0.3s; }
        .tooltip:hover .tooltiptext { visibility: visible; opacity: 1; }
    </style>
    """

    # JavaScript for tab functionality
    script = """
    <script>
    function openTab(evt, tabName) {
        var i, tabcontent, tablinks;
        tabcontent = document.getElementsByClassName("tabcontent");
        for (i = 0; i < tabcontent.length; i++) {
            tabcontent[i].style.display = "none";
        }
        tablinks = document.getElementsByClassName("tablinks");
        for (i = 0; i < tablinks.length; i++) {
            tablinks[i].className = tablinks[i].className.replace(" active", "");
        }
        document.getElementById(tabName).style.display = "block";
        evt.currentTarget.className += " active";
    }
    </script>
    """

    # JavaScript for dynamic tooltip positioning
    dynamic_script = """
    <script>
    function adjustTooltipPosition(event) {
        var tooltip = event.target.querySelector('.tooltiptext');
        var rect = tooltip.getBoundingClientRect();
        var topSpace = rect.top;
        var bottomSpace = window.innerHeight - rect.bottom;

        if (topSpace < bottomSpace) {
            tooltip.style.bottom = '125%';
        } else {
            tooltip.style.bottom = 'auto';
            tooltip.style.top = '125%';
        }
    }
    </script>
    """

    # HTML for tabs
    tab_html = '<div class="tab">'
    content_html = ""
    for i, (title, df) in enumerate(bpa_dict.items()):
        if df.shape[0] == 0:
            continue

        tab_id = f"tab{i}"
        active_class = ""
        if i == 0:
            active_class = "active"

        summary = " + ".join(
            [f"{idx} ({v})" for idx, v in df["Severity"].value_counts().items()]
        )
        tab_html += f'<button class="tablinks {active_class}" onclick="openTab(event, \'{tab_id}\')"><b>{title}</b><br/>{summary}</button>'
        content_html += f'<div id="{tab_id}" class="tabcontent {active_class}">'

        # Adding tooltip for Rule Name using Description column
        content_html += '<table border="1">'
        content_html += "<tr><th>Rule Name</th><th>Object Type</th><th>Object Name</th><th>Severity</th></tr>"
        for _, row in df.iterrows():
            content_html += "<tr>"
            if pd.notnull(row["URL"]):
                content_html += f'<td class="tooltip" onmouseover="adjustTooltipPosition(event)"><a href="{row["URL"]}">{row["Rule Name"]}</a><span class="tooltiptext">{row["Description"]}</span></td>'
            elif pd.notnull(row["Description"]):
                content_html += f'<td class="tooltip" onmouseover="adjustTooltipPosition(event)">{row["Rule Name"]}<span class="tooltiptext">{row["Description"]}</span></td>'
            else:
                content_html += f'<td>{row["Rule Name"]}</td>'
            content_html += f'<td>{row["Object Type"]}</td>'
            content_html += f'<td>{row["Object Name"]}</td>'
            content_html += f'<td>{row["Severity"]}</td>'
            content_html += "</tr>"
        content_html += "</table>"

        content_html += "</div>"
    tab_html += "</div>"

    # Display the tabs, tab contents, and run the script
    return display(HTML(styles + tab_html + content_html + script))



import sempy.fabric as fabric
from sempy_labs._helper_functions import resolve_dataset_id, resolve_report_id
from typing import Optional, List
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy.fabric.exceptions import FabricHTTPException


@log
def report_rebind(
    report: str | List[str],
    dataset: str,
    report_workspace: Optional[str] = None,
    dataset_workspace: Optional[str] = None,
):
    """
    Rebinds a report to a semantic model.

    Parameters
    ----------
    report : str | List[str]
        Name(s) of the Power BI report(s).
    dataset : str
        Name of the semantic model.
    report_workspace : str, default=None
        The name of the Fabric workspace in which the report resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    dataset_workspace : str, default=None
        The name of the Fabric workspace in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    if report_workspace is None:
        report_workspace_id = fabric.get_workspace_id()
        report_workspace = fabric.resolve_workspace_name(report_workspace_id)
    else:
        report_workspace_id = fabric.resolve_workspace_id(report_workspace)
    if dataset_workspace is None:
        dataset_workspace = report_workspace

    client = fabric.PowerBIRestClient()

    if isinstance(report, str):
        report = [report]

    for rpt in report:
        reportId = resolve_report_id(report=rpt, workspace=report_workspace)
        datasetId = resolve_dataset_id(dataset=dataset, workspace=dataset_workspace)

        # Prepare API
        request_body = {"datasetId": datasetId}

        response = client.post(
            f"/v1.0/myorg/groups/{report_workspace_id}/reports/{reportId}/Rebind",
            json=request_body,
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)
        print(
            f"{icons.green_dot} The '{rpt}' report has been successfully rebinded to the '{dataset}' semantic model."
        )


@log
def report_rebind_all(
    dataset: str,
    new_dataset: str,
    dataset_workspace: Optional[str] = None,
    new_dataset_workpace: Optional[str] = None,
    report_workspace: Optional[str | List[str]] = None,
):
    """
    Rebinds all reports in a workspace which are bound to a specific semantic model to a new semantic model.

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    dataset : str
        Name of the semantic model currently binded to the reports.
    new_dataset : str
        Name of the semantic model to rebind to the reports.
    dataset_workspace : str, default=None
        The name of the Fabric workspace in which the original semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataset_workspace : str, default=None
        The name of the Fabric workspace in which the new semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    report_workspace : str | List[str], default=None
        The name(s) of the Fabric workspace(s) in which the report(s) reside(s).
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    dataset_workspace = fabric.resolve_workspace_name()

    if new_dataset_workpace is None:
        new_dataset_workpace = dataset_workspace

    if report_workspace is None:
        report_workspace = dataset_workspace

    if isinstance(report_workspace, str):
        report_workspace = [report_workspace]

    datasetId = resolve_dataset_id(dataset, dataset_workspace)

    for rw in report_workspace:
        dfRep = fabric.list_reports(workspace=rw)
        dfRep_filt = dfRep[dfRep["Dataset Id"] == datasetId]
        for i, r in dfRep_filt.iterrows():
            rptName = r["Name"]
            report_rebind(
                report=rptName,
                dataset=new_dataset,
                report_workspace=rw,
                dataset_workspace=new_dataset_workpace,
            )



import sempy
import sempy.fabric as fabric
import pandas as pd
from sempy_labs.lakehouse._get_lakehouse_columns import get_lakehouse_columns
from sempy_labs.tom import connect_semantic_model
from sempy_labs._helper_functions import (
    format_dax_object_name,
    resolve_lakehouse_name,
    get_direct_lake_sql_endpoint,
)
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def direct_lake_schema_sync(
    dataset: str,
    workspace: Optional[str] = None,
    add_to_model: Optional[bool] = False,
    lakehouse: Optional[str] = None,
    lakehouse_workspace: Optional[str] = None,
):
    """
    Shows/adds columns which exist in the lakehouse but do not exist in the semantic model (only for tables in the semantic model).

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    add_to_model : bool, default=False
        If set to True, columns which exist in the lakehouse but do not exist in the semantic model are added to the semantic model. No new tables are added.
    lakehouse : str, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str, default=None
        The Fabric workspace used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM
    import System

    workspace = fabric.resolve_workspace_name(workspace)

    if lakehouse_workspace is None:
        lakehouse_workspace = workspace

    if lakehouse is None:
        lakehouse_id = fabric.get_lakehouse_id()
        lakehouse = resolve_lakehouse_name(lakehouse_id, lakehouse_workspace)

    sqlEndpointId = get_direct_lake_sql_endpoint(dataset, workspace)

    dfI = fabric.list_items(workspace=lakehouse_workspace, type="SQLEndpoint")
    dfI_filt = dfI[(dfI["Id"] == sqlEndpointId)]

    if len(dfI_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The SQL Endpoint in the '{dataset}' semantic model in the '{workspace} workspace does not point to the "
            f"'{lakehouse}' lakehouse in the '{lakehouse_workspace}' workspace as specified."
        )

    dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    dfP_filt = dfP[dfP["Source Type"] == "Entity"]
    dfC = fabric.list_columns(dataset=dataset, workspace=workspace)
    dfC_filt = dfC[dfC["Table Name"].isin(dfP_filt["Table Name"].values)]
    dfC_filt = pd.merge(
        dfC_filt, dfP_filt[["Table Name", "Query"]], on="Table Name", how="left"
    )
    dfC_filt["Column Object"] = format_dax_object_name(
        dfC_filt["Query"], dfC_filt["Source"]
    )

    lc = get_lakehouse_columns(lakehouse, lakehouse_workspace)
    lc_filt = lc[lc["Table Name"].isin(dfP_filt["Query"].values)]

    with connect_semantic_model(
        dataset=dataset, readonly=False, workspace=workspace
    ) as tom:

        for i, r in lc_filt.iterrows():
            lakeTName = r["Table Name"]
            lakeCName = r["Column Name"]
            fullColName = r["Full Column Name"]
            dType = r["Data Type"]

            if fullColName not in dfC_filt["Column Object"].values:
                dfL = dfP_filt[dfP_filt["Query"] == lakeTName]
                tName = dfL["Table Name"].iloc[0]
                if add_to_model:
                    col = TOM.DataColumn()
                    col.Name = lakeCName
                    col.SourceColumn = lakeCName
                    dt = icons.data_type_mapping.get(dType)
                    try:
                        col.DataType = System.Enum.Parse(TOM.DataType, dt)
                    except Exception as e:
                        raise ValueError(
                            f"{icons.red_dot} Failed to map '{dType}' data type to the semantic model data types."
                        ) from e

                    tom.model.Tables[tName].Columns.Add(col)
                    print(
                        f"{icons.green_dot} The '{lakeCName}' column has been added to the '{tName}' table as a '{dt}' "
                        f"data type within the '{dataset}' semantic model within the '{workspace}' workspace."
                    )
                else:
                    print(
                        f"{icons.yellow_dot} The {fullColName} column exists in the lakehouse but not in the '{tName}' table in the '{dataset}' semantic model within the '{workspace}' workspace."
                    )



import sempy
import sempy.fabric as fabric
import re
import datetime
import time
from sempy_labs._list_functions import list_tables
from sempy_labs._helper_functions import create_relationship_name
from sempy_labs.tom import connect_semantic_model
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def migrate_model_objects_to_semantic_model(
    dataset: str,
    new_dataset: str,
    workspace: Optional[str] = None,
    new_dataset_workspace: Optional[str] = None,
):
    """
    Adds the rest of the model objects (besides tables/columns) and their properties to a Direct Lake semantic model based on an import/DirectQuery semantic model.

    Parameters
    ----------
    dataset : str
        Name of the import/DirectQuery semantic model.
    new_dataset : str
        Name of the Direct Lake semantic model.
    workspace : str, default=None
        The Fabric workspace name in which the import/DirectQuery semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataset_workspace : str
        The Fabric workspace name in which the Direct Lake semantic model will be created.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM
    import System

    workspace = fabric.resolve_workspace_name(workspace)

    if new_dataset_workspace is None:
        new_dataset_workspace = workspace

    dfT = list_tables(dataset, workspace)
    dfC = fabric.list_columns(dataset=dataset, workspace=workspace)
    dfM = fabric.list_measures(dataset=dataset, workspace=workspace)
    dfR = fabric.list_relationships(dataset=dataset, workspace=workspace)
    dfRole = fabric.get_roles(dataset=dataset, workspace=workspace)
    dfRLS = fabric.get_row_level_security_permissions(
        dataset=dataset, workspace=workspace
    )
    dfCI = fabric.list_calculation_items(dataset=dataset, workspace=workspace)
    dfP = fabric.list_perspectives(dataset=dataset, workspace=workspace)
    dfTranslation = fabric.list_translations(dataset=dataset, workspace=workspace)
    dfH = fabric.list_hierarchies(dataset=dataset, workspace=workspace)
    dfPar = fabric.list_partitions(dataset=dataset, workspace=workspace)

    dfP_cc = dfPar[(dfPar["Source Type"] == "Calculated")]
    dfP_fp = dfP_cc[dfP_cc["Query"].str.contains("NAMEOF")]

    print(f"{icons.in_progress} Updating '{new_dataset}' based on '{dataset}'...")
    start_time = datetime.datetime.now()
    timeout = datetime.timedelta(minutes=1)
    success = False

    while not success:
        try:
            with connect_semantic_model(
                dataset=new_dataset, readonly=False, workspace=new_dataset_workspace
            ) as tom:
                success = True

                isDirectLake = any(
                    str(p.Mode) == "DirectLake"
                    for t in tom.model.Tables
                    for p in t.Partitions
                )

                print(f"\n{icons.in_progress} Updating table properties...")
                for t in tom.model.Tables:
                    t.IsHidden = bool(dfT.loc[dfT["Name"] == t.Name, "Hidden"].iloc[0])
                    t.Description = dfT.loc[dfT["Name"] == t.Name, "Description"].iloc[
                        0
                    ]
                    t.DataCategory = dfT.loc[
                        dfT["Name"] == t.Name, "Data Category"
                    ].iloc[0]

                    print(
                        f"{icons.green_dot} The '{t.Name}' table's properties have been updated."
                    )

                print(f"\n{icons.in_progress} Updating column properties...")
                for t in tom.model.Tables:
                    if (
                        t.Name not in dfP_fp["Table Name"].values
                    ):  # do not include field parameters
                        dfT_filtered = dfT[dfT["Name"] == t.Name]
                        tType = dfT_filtered["Type"].iloc[0]
                        for c in t.Columns:
                            if not c.Name.startswith("RowNumber-"):
                                dfC_filt = dfC[
                                    (dfC["Table Name"] == t.Name)
                                    & (dfC["Column Name"] == c.Name)
                                ]
                                cName = dfC_filt["Column Name"].iloc[0]
                                c.Name = cName
                                if tType == "Table":
                                    c.SourceColumn = cName.replace(" ", "_")
                                c.IsHidden = bool(dfC_filt["Hidden"].iloc[0])
                                c.DataType = System.Enum.Parse(
                                    TOM.DataType, dfC_filt["Data Type"].iloc[0]
                                )
                                c.DisplayFolder = dfC_filt["Display Folder"].iloc[0]
                                c.FormatString = dfC_filt["Format String"].iloc[0]
                                c.SummarizeBy = System.Enum.Parse(
                                    TOM.AggregateFunction,
                                    dfC_filt["Summarize By"].iloc[0],
                                )
                                c.DataCategory = dfC_filt["Data Category"].iloc[0]
                                c.IsKey = bool(dfC_filt["Key"].iloc[0])
                                sbc = dfC_filt["Sort By Column"].iloc[0]

                                if sbc is not None:
                                    if any(
                                        o.Name == sbc and o.Parent.Name == c.Parent.Name
                                        for o in tom.all_columns()
                                    ):
                                        c.SortByColumn = tom.model.Tables[
                                            t.Name
                                        ].Columns[sbc]
                                    else:
                                        print(
                                            f"{icons.red_dot} Failed to create '{sbc}' as a Sort By Column for the '{c.Name}' in the '{t.Name}' table."
                                        )
                                print(
                                    f"{icons.green_dot} The '{t.Name}'[{c.Name}] column's properties have been updated."
                                )

                print(f"\n{icons.in_progress} Creating hierarchies...")
                dfH_grouped = (
                    dfH.groupby(
                        [
                            "Table Name",
                            "Hierarchy Name",
                            "Hierarchy Hidden",
                            "Hierarchy Description",
                        ]
                    )
                    .agg({"Level Name": list, "Column Name": list})
                    .reset_index()
                )

                for i, r in dfH_grouped.iterrows():
                    tName = r["Table Name"]
                    hName = r["Hierarchy Name"]
                    hDesc = r["Hierarchy Description"]
                    hHid = bool(r["Hierarchy Hidden"])
                    cols = r["Column Name"]
                    lvls = r["Level Name"]

                    if not any(
                        t.Name == tName and h.Name == hName
                        for t in tom.model.Tables
                        for h in t.Hierarchies
                    ):
                        tom.add_hierarchy(
                            table_name=tName,
                            hierarchy_name=hName,
                            hierarchy_description=hDesc,
                            hierarchy_hidden=hHid,
                            columns=cols,
                            levels=lvls,
                        )
                        print(
                            f"{icons.green_dot} The '{hName}' hierarchy has been added."
                        )

                print(f"\n{icons.in_progress} Creating measures...")
                for i, r in dfM.iterrows():
                    tName = r["Table Name"]
                    mName = r["Measure Name"]
                    mExpr = r["Measure Expression"]
                    mHidden = bool(r["Measure Hidden"])
                    mDF = r["Measure Display Folder"]
                    mDesc = r["Measure Description"]
                    mFS = r["Format String"]

                    if not any(m.Name == mName for m in tom.all_measures()):
                        tom.add_measure(
                            table_name=tName,
                            measure_name=mName,
                            expression=mExpr,
                            hidden=mHidden,
                            display_folder=mDF,
                            description=mDesc,
                            format_string=mFS,
                        )
                        print(
                            f"{icons.green_dot} The '{mName}' measure has been added."
                        )
                print(f"\n{icons.in_progress} Creating calculation groups...")
                for cgName in dfCI["Calculation Group Name"].unique():

                    isHidden = bool(
                        dfCI.loc[
                            (dfCI["Calculation Group Name"] == cgName), "Hidden"
                        ].iloc[0]
                    )
                    prec = int(
                        dfCI.loc[
                            (dfCI["Calculation Group Name"] == cgName), "Precedence"
                        ].iloc[0]
                    )
                    desc = dfCI.loc[
                        (dfCI["Calculation Group Name"] == cgName), "Description"
                    ].iloc[0]

                    if not any(t.Name == cgName for t in tom.model.Tables):
                        tom.add_calculation_group(
                            name=cgName,
                            description=desc,
                            precedence=prec,
                            hidden=isHidden,
                        )
                        print(
                            f"{icons.green_dot} The '{cgName}' calculation group has been added."
                        )
                        tom.model.DiscourageImplicitMeasures = True

                        # print(
                        #    f"\n{icons.in_progress} Updating calculation group column names..."
                        # )
                        dfC_filt = dfC[
                            (dfC["Table Name"] == cgName) & (dfC["Hidden"] == False)
                        ]
                        colName = dfC_filt["Column Name"].iloc[0]
                        tom.model.Tables[cgName].Columns["Name"].Name = colName

                    calcItems = dfCI.loc[
                        dfCI["Calculation Group Name"] == cgName,
                        "Calculation Item Name",
                    ].unique()

                    print(f"\n{icons.in_progress} Creating calculation items...")
                    for calcItem in calcItems:
                        ordinal = int(
                            dfCI.loc[
                                (dfCI["Calculation Group Name"] == cgName)
                                & (dfCI["Calculation Item Name"] == calcItem),
                                "Ordinal",
                            ].iloc[0]
                        )
                        expr = dfCI.loc[
                            (dfCI["Calculation Group Name"] == cgName)
                            & (dfCI["Calculation Item Name"] == calcItem),
                            "Expression",
                        ].iloc[0]
                        fse = dfCI.loc[
                            (dfCI["Calculation Group Name"] == cgName)
                            & (dfCI["Calculation Item Name"] == calcItem),
                            "Format String Expression",
                        ].iloc[0]

                        if not any(
                            ci.CalculationGroup.Parent.Name == cgName
                            and ci.Name == calcItem
                            for ci in tom.all_calculation_items()
                        ):
                            tom.add_calculation_item(
                                table_name=cgName,
                                calculation_item_name=calcItem,
                                expression=expr,
                                format_string_expression=fse,
                                ordinal=ordinal,
                            )
                            print(
                                f"{icons.green_dot} The '{calcItem}' has been added to the '{cgName}' calculation group."
                            )

                print(f"\n{icons.in_progress} Creating relationships...")
                for index, row in dfR.iterrows():
                    fromTable = row["From Table"]
                    fromColumn = row["From Column"]
                    toTable = row["To Table"]
                    toColumn = row["To Column"]
                    isActive = row["Active"]
                    cfb = row["Cross Filtering Behavior"]
                    sfb = row["Security Filtering Behavior"]
                    rori = row["Rely On Referential Integrity"]
                    mult = row["Multiplicity"]

                    card_mapping = {"m": "Many", "1": "One", "0": "None"}

                    fromCard = card_mapping.get(mult[0])
                    toCard = card_mapping.get(mult[-1])

                    relName = create_relationship_name(
                        fromTable, fromColumn, toTable, toColumn
                    )

                    if any(
                        r.FromTable.Name == fromTable
                        and r.FromColumn.Name == fromColumn
                        and r.ToTable.Name == toTable
                        and r.ToColumn.Name == toColumn
                        for r in tom.model.Relationships
                    ):
                        print(
                            f"{icons.yellow_dot} {relName} already exists as a relationship in the semantic model."
                        )
                    elif isDirectLake and any(
                        r.FromTable.Name == fromTable
                        and r.FromColumn.Name == fromColumn
                        and r.ToTable.Name == toTable
                        and r.ToColumn.Name == toColumn
                        and (
                            r.FromColumn.DataType == "DateTime"
                            or r.ToColumn.DataType == "DateTime"
                        )
                        for r in tom.model.Relationships
                    ):
                        print(
                            f"{icons.yellow_dot} {relName} was not created since relationships based on DateTime columns are not supported."
                        )
                    elif isDirectLake and any(
                        r.FromTable.Name == fromTable
                        and r.FromColumn.Name == fromColumn
                        and r.ToTable.Name == toTable
                        and r.ToColumn.Name == toColumn
                        and (r.FromColumn.DataType != r.ToColumn.DataType)
                        for r in tom.model.Relationships
                    ):
                        print(
                            f"{icons.yellow_dot} {relName} was not created since columns used in a relationship must have the same data type."
                        )
                    else:
                        if not any(
                            r.FromTable.Name == fromTable
                            and r.FromColumn.Name == fromColumn
                            and r.ToTable.Name == toTable
                            and r.ToColumn.Name == toColumn
                            for r in tom.model.Relationships
                        ):
                            tom.add_relationship(
                                from_table=fromTable,
                                from_column=fromColumn,
                                to_table=toTable,
                                to_column=toColumn,
                                from_cardinality=fromCard,
                                to_cardinality=toCard,
                                cross_filtering_behavior=cfb,
                                security_filtering_behavior=sfb,
                                rely_on_referential_integrity=rori,
                                is_active=isActive,
                            )
                            print(
                                f"{icons.green_dot} The {relName} relationship has been added."
                            )
                        else:
                            print(
                                f"{icons.red_dot} The {relName} relationship was not added."
                            )

                print(f"\n{icons.in_progress} Creating roles...")
                for index, row in dfRole.iterrows():
                    roleName = row["Role"]
                    roleDesc = row["Description"]
                    modPerm = row["Model Permission"]

                    if not any(r.Name == roleName for r in tom.model.Roles):
                        tom.add_role(
                            role_name=roleName,
                            model_permission=modPerm,
                            description=roleDesc,
                        )
                        print(
                            f"{icons.green_dot} The '{roleName}' role has been added."
                        )

                print(f"\n{icons.in_progress} Creating row level security...")
                for index, row in dfRLS.iterrows():
                    roleName = row["Role"]
                    tName = row["Table"]
                    expr = row["Filter Expression"]

                    if any(t.Name == tName for t in tom.model.Tables):
                        tom.set_rls(
                            role_name=roleName, table_name=tName, filter_expression=expr
                        )
                        print(
                            f"{icons.green_dot} Row level security for the '{tName}' table within the '{roleName}' role has been set."
                        )
                    else:
                        print(
                            f"{icons.red_dot} Row level security for the '{tName}' table within the '{roleName}' role was not set."
                        )

                print(f"\n{icons.in_progress} Creating perspectives...")
                for pName in dfP["Perspective Name"].unique():

                    if not any(p.Name == pName for p in tom.model.Perspectives):
                        tom.add_perspective(perspective_name=pName)
                        print(
                            f"{icons.green_dot} The '{pName}' perspective has been added."
                        )

                print(f"\n{icons.in_progress} Adding objects to perspectives...")
                for index, row in dfP.iterrows():
                    pName = row["Perspective Name"]
                    tName = row["Table Name"]
                    oName = row["Object Name"]
                    oType = row["Object Type"]
                    tType = dfT.loc[(dfT["Name"] == tName), "Type"].iloc[0]

                    try:
                        if oType == "Table":
                            tom.add_to_perspective(
                                object=tom.model.Tables[tName], perspective_name=pName
                            )
                        elif oType == "Column":
                            tom.add_to_perspective(
                                object=tom.model.Tables[tName].Columns[oName],
                                perspective_name=pName,
                            )
                        elif oType == "Measure":
                            tom.add_to_perspective(
                                object=tom.model.Tables[tName].Measures[oName],
                                perspective_name=pName,
                            )
                        elif oType == "Hierarchy":
                            tom.add_to_perspective(
                                object=tom.model.Tables[tName].Hierarchies[oName],
                                perspective_name=pName,
                            )
                    except Exception:
                        pass

                print(f"\n{icons.in_progress} Creating translation languages...")
                for trName in dfTranslation["Culture Name"].unique():
                    if not any(c.Name == trName for c in tom.model.Cultures):
                        tom.add_translation(trName)
                        print(
                            f"{icons.green_dot} The '{trName}' translation language has been added."
                        )

                print(f"\n{icons.in_progress} Creating translation values...")
                for index, row in dfTranslation.iterrows():
                    trName = row["Culture Name"]
                    tName = row["Table Name"]
                    oName = row["Object Name"]
                    oType = row["Object Type"]
                    translation = row["Translation"]
                    prop = row["Property"]

                    if prop == "Caption":
                        prop = "Name"
                    elif prop == "DisplayFolder":
                        prop = "Display Folder"

                    try:
                        if oType == "Table":
                            tom.set_translation(
                                object=tom.model.Tables[tName],
                                language=trName,
                                property=prop,
                                value=translation,
                            )
                        elif oType == "Column":
                            tom.set_translation(
                                object=tom.model.Tables[tName].Columns[oName],
                                language=trName,
                                property=prop,
                                value=translation,
                            )
                        elif oType == "Measure":
                            tom.set_translation(
                                object=tom.model.Tables[tName].Measures[oName],
                                language=trName,
                                property=prop,
                                value=translation,
                            )
                        elif oType == "Hierarchy":
                            tom.set_translation(
                                object=tom.model.Tables[tName].Hierarchies[oName],
                                language=trName,
                                property=prop,
                                value=translation,
                            )
                        elif oType == "Level":

                            pattern = r"\[([^]]+)\]"
                            matches = re.findall(pattern, oName)
                            lName = matches[0]

                            pattern = r"'([^']+)'"
                            matches = re.findall(pattern, oName)
                            hName = matches[0]
                            tom.set_translation(
                                object=tom.model.Tables[tName]
                                .Hierarchies[hName]
                                .Levels[lName],
                                language=trName,
                                property=prop,
                                value=translation,
                            )
                    except Exception:
                        pass

                print(
                    f"\n{icons.green_dot} Migration of objects from '{dataset}' -> '{new_dataset}' is complete."
                )

        except Exception:
            if datetime.datetime.now() - start_time > timeout:
                break
            time.sleep(1)



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import resolve_dataset_id
from typing import Optional
import sempy_labs._icons as icons
from sempy.fabric.exceptions import FabricHTTPException


def qso_sync(dataset: str, workspace: Optional[str] = None):
    """
    Triggers a query scale-out sync of read-only replicas for the specified dataset from the specified workspace.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/trigger-query-scale-out-sync-in-group

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    dataset_id = resolve_dataset_id(dataset, workspace)

    client = fabric.PowerBIRestClient()
    response = client.post(
        f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/queryScaleOut/sync"
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)
    print(
        f"{icons.green_dot} QSO sync initiated for the '{dataset}' semantic model within the '{workspace}' workspace."
    )


def qso_sync_status(dataset: str, workspace: Optional[str] = None):
    """
    Returns the query scale-out sync status for the specified dataset from the specified workspace.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/get-query-scale-out-sync-status-in-group

    df = pd.DataFrame(
        columns=[
            "Scale Out Status",
            "Sync Start Time",
            "Sync End Time",
            "Commit Version",
            "Commit Timestamp",
            "Target Sync Version",
            "Target Sync Timestamp",
            "Trigger Reason",
            "Min Active Read Version",
            "Min Active Read Timestamp",
        ]
    )
    dfRep = pd.DataFrame(
        columns=["Replica ID", "Replica Type", "Replica Version", "Replica Timestamp"]
    )

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    dataset_id = resolve_dataset_id(dataset, workspace)

    client = fabric.PowerBIRestClient()
    response = client.get(
        f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/queryScaleOut/syncStatus"
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)

    o = response.json()
    sos = o["scaleOutStatus"]

    if sos == "Enabled":
        new_data = {
            "Scale Out Status": o["scaleOutStatus"],
            "Sync Start Time": o["syncStartTime"],
            "Sync End Time": o["syncEndTime"],
            "Commit Version": o["commitVersion"],
            "Commit Timestamp": o["commitTimestamp"],
            "Target Sync Version": o["targetSyncVersion"],
            "Target Sync Timestamp": o["targetSyncTimestamp"],
            "Trigger Reason": o["triggerReason"],
            "Min Active Read Version": o["minActiveReadVersion"],
            "Min Active Read Timestamp": o["minActiveReadTimestamp"],
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

        for r in o["scaleOutReplicas"]:
            new_data = {
                "Replica ID": r["replicaId"],
                "Replica Type": r["replicaType"],
                "Replica Version": str(r["replicaVersion"]),
                "Replica Timestamp": r["replicaTimestamp"],
            }
            dfRep = pd.concat(
                [dfRep, pd.DataFrame(new_data, index=[0])], ignore_index=True
            )

        df["Sync Start Time"] = pd.to_datetime(df["Sync Start Time"])
        df["Sync End Time"] = pd.to_datetime(df["Sync End Time"])
        df["Commit Timestamp"] = pd.to_datetime(df["Commit Timestamp"])
        df["Target Sync Timestamp"] = pd.to_datetime(df["Target Sync Timestamp"])
        df["Min Active Read Timestamp"] = pd.to_datetime(
            df["Min Active Read Timestamp"]
        )
        dfRep["Replica Timestamp"] = pd.to_datetime(dfRep["Replica Timestamp"])
        df["Commit Version"] = df["Commit Version"].astype("int")
        df["Target Sync Version"] = df["Target Sync Version"].astype("int")
        df["Min Active Read Version"] = df["Min Active Read Version"].astype("int")

        return df, dfRep
    else:
        print(f"{sos}\n\n")
        return df, dfRep


def disable_qso(dataset: str, workspace: Optional[str] = None):
    """
    Sets the max read-only replicas to 0, disabling query scale out.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    dataset_id = resolve_dataset_id(dataset, workspace)

    request_body = {"queryScaleOutSettings": {"maxReadOnlyReplicas": "0"}}

    client = fabric.PowerBIRestClient()
    response = client.patch(
        f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}", json=request_body
    )
    if response.status_code != 200:
        raise FabricHTTPException(response)

    df = list_qso_settings(dataset=dataset, workspace=workspace)
    print(
        f"{icons.green_dot} Query scale out has been disabled for the '{dataset}' semantic model within the '{workspace}' workspace."
    )

    return df


def set_qso(
    dataset: str,
    auto_sync: Optional[bool] = True,
    max_read_only_replicas: Optional[int] = -1,
    workspace: Optional[str] = None,
):
    """
    Sets the query scale out settings for a semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    auto_sync : bool, default=True
        Whether the semantic model automatically syncs read-only replicas.
    max_read_only_replicas : int, default=-1
        To enable semantic model scale-out, set max_read_only_replicas to -1, or any non-0 value. A value of -1 allows Power BI to create as many read-only replicas as your Power BI capacity supports. You can also explicitly set the replica count to a value lower than that of the capacity maximum. Setting max_read_only_replicas to -1 is recommended.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/update-dataset-in-group

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    dataset_id = resolve_dataset_id(dataset, workspace)

    if max_read_only_replicas == 0:
        disable_qso(dataset=dataset, workspace=workspace)
        return

    request_body = {
        "queryScaleOutSettings": {
            "autoSyncReadOnlyReplicas": auto_sync,
            "maxReadOnlyReplicas": str(max_read_only_replicas),
        }
    }

    ssm = set_semantic_model_storage_format(
        dataset=dataset, storage_format="Large", workspace=workspace
    )
    if ssm == 200:
        client = fabric.PowerBIRestClient()
        response = client.patch(
            f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}",
            json=request_body,
        )
        if response.status_code != 200:
            raise FabricHTTPException(response)

        df = list_qso_settings(dataset=dataset, workspace=workspace)
        print(
            f"{icons.green_dot} Query scale out has been set on the '{dataset}' semantic model within the '{workspace}' workspace."
        )
        return df
    else:
        raise ValueError(
            f"{icons.red_dot} Failed to set the '{dataset}' semantic model within the '{workspace}' workspace to large semantic model storage format. This is a prerequisite for enabling Query Scale Out.\n\"https://learn.microsoft.com/power-bi/enterprise/service-premium-scale-out#prerequisites\""
        )


def set_semantic_model_storage_format(
    dataset: str, storage_format: str, workspace: Optional[str] = None
):
    """
    Sets the semantic model storage format.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    storage_format : str
        The storage format for the semantic model. Valid options: 'Large', 'Small'.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    dataset_id = resolve_dataset_id(dataset, workspace)

    storage_format = storage_format.capitalize()

    if storage_format == "Abf":
        storage_format = "Small"
    elif storage_format.startswith("Premium"):
        storage_format = "Large"

    storageFormats = ["Small", "Large"]

    if storage_format == "Large":
        request_body = {"targetStorageMode": "PremiumFiles"}
    elif storage_format == "Small":
        request_body = {"targetStorageMode": "Abf"}
    else:
        raise ValueError(
            f"{icons.red_dot} Invalid storage format value. Valid options: {storageFormats}."
        )

    client = fabric.PowerBIRestClient()
    response = client.patch(
        f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}", json=request_body
    )
    if response.status_code != 200:
        raise FabricHTTPException(response)
    print(f"{icons.green_dot} Semantic model storage format set to '{storage_format}'.")

    return response.status_code


def list_qso_settings(dataset: Optional[str] = None, workspace: Optional[str] = None):
    """
    Shows the query scale out settings for a semantic model (or all semantic models within a workspace).

    Parameters
    ----------
    dataset : str, default=None
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the query scale out settings.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    if dataset is not None:
        dataset_id = resolve_dataset_id(dataset, workspace)

    workspace_id = fabric.get_workspace_id()
    df = pd.DataFrame(
        columns=[
            "Dataset Id",
            "Dataset Name",
            "Storage Mode",
            "QSO Auto Sync Enabled",
            "QSO Max Read Only Replicas",
        ]
    )
    client = fabric.PowerBIRestClient()
    response = client.get(f"/v1.0/myorg/groups/{workspace_id}/datasets")
    for v in response.json()["value"]:
        tsm = v.get("targetStorageMode")
        if tsm == "Abf":
            sm = "Small"
        else:
            sm = "Large"
        new_data = {
            "Dataset Id": v.get("id"),
            "Dataset Name": v.get("name"),
            "Storage Mode": sm,
            "QSO Auto Sync Enabled": v.get("queryScaleOutSettings", {}).get(
                "autoSyncReadOnlyReplicas"
            ),
            "QSO Max Read Only Replicas": v.get("queryScaleOutSettings", {}).get(
                "maxReadOnlyReplicas"
            ),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    df["QSO Auto Sync Enabled"] = df["QSO Auto Sync Enabled"].astype("bool")
    df["QSO Max Read Only Replicas"] = df["QSO Max Read Only Replicas"].astype("int")

    if dataset is not None:
        df = df[df["Dataset Id"] == dataset_id]

    return df


def set_workspace_default_storage_format(
    storage_format: str, workspace: Optional[str] = None
):
    """
    Sets the default storage format for semantic models within a workspace.

    Parameters
    ----------
    storage_format : str
        The storage format for the semantic model. Valid options: 'Large', 'Small'.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/groups/update-group#defaultdatasetstorageformat

    storageFormats = ["Small", "Large"]

    storage_format = storage_format.capitalize()

    if storage_format not in storageFormats:
        print(
            f"Invalid storage format. Please choose from these options: {storageFormats}."
        )

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    request_body = {"name": workspace, "defaultDatasetStorageFormat": storage_format}

    client = fabric.PowerBIRestClient()
    response = client.patch(f"/v1.0/myorg/groups/{workspace_id}", json=request_body)

    if response.status_code != 200:
        raise FabricHTTPException(response)
    print(
        f"{icons.green_dot} The default storage format for the '{workspace}' workspace has been updated to '{storage_format}."
    )



from sempy_labs._clear_cache import clear_cache

# from sempy_labs._connections import (
# create_connection_cloud,
# create_connection_vnet,
# create_connection_on_prem
# )
from sempy_labs._dax import evaluate_dax_impersonation
from sempy_labs._generate_semantic_model import (
    create_blank_semantic_model,
    create_semantic_model_from_bim,
    deploy_semantic_model,
    get_semantic_model_bim,
)
from sempy_labs._list_functions import (
    delete_custom_pool,
    list_semantic_model_objects,
    list_shortcuts,
    get_object_level_security,
    list_capacities,
    # list_annotations,
    # list_columns,
    list_dashboards,
    list_dataflow_storage_accounts,
    # list_datamarts,
    # list_datapipelines,
    # list_eventstreams,
    # list_kpis,
    # list_kqldatabases,
    # list_kqlquerysets,
    list_lakehouses,
    # list_mirroredwarehouses,
    # list_mlexperiments,
    # list_mlmodels,
    # list_relationships,
    # list_sqlendpoints,
    # list_tables,
    list_warehouses,
    list_workspace_role_assignments,
    create_warehouse,
    update_item,
    list_custom_pools,
    create_custom_pool,
    update_custom_pool,
    assign_workspace_to_capacity,
    unassign_workspace_from_capacity,
    get_spark_settings,
    update_spark_settings,
    add_user_to_workspace,
    delete_user_from_workspace,
    update_workspace_user,
    list_workspace_users,
    assign_workspace_to_dataflow_storage,
)

from sempy_labs._helper_functions import (
    create_abfss_path,
    format_dax_object_name,
    create_relationship_name,
    save_as_delta_table,
    generate_embedded_filter,
    get_direct_lake_sql_endpoint,
    resolve_lakehouse_id,
    resolve_lakehouse_name,
    resolve_dataset_id,
    resolve_dataset_name,
    resolve_report_id,
    resolve_report_name,
    #  language_validate
)

#  from sempy_labs._model_auto_build import model_auto_build
from sempy_labs._model_bpa import run_model_bpa
from sempy_labs._model_bpa_rules import model_bpa_rules
from sempy_labs._model_dependencies import (
    measure_dependency_tree,
    get_measure_dependencies,
    get_model_calc_dependencies,
)
from sempy_labs._one_lake_integration import (
    export_model_to_onelake,
)
from sempy_labs._query_scale_out import (
    qso_sync,
    qso_sync_status,
    set_qso,
    list_qso_settings,
    disable_qso,
    set_semantic_model_storage_format,
    set_workspace_default_storage_format,
)
from sempy_labs._refresh_semantic_model import (
    refresh_semantic_model,
    cancel_dataset_refresh,
)
from sempy_labs._translations import translate_semantic_model
from sempy_labs._vertipaq import (
    vertipaq_analyzer,
    # visualize_vertipaq,
    import_vertipaq_analyzer,
)

__all__ = [
    "delete_custom_pool",
    "clear_cache",
    # create_connection_cloud,
    # create_connection_vnet,
    # create_connection_on_prem,
    "evaluate_dax_impersonation",
    "create_blank_semantic_model",
    "create_semantic_model_from_bim",
    "deploy_semantic_model",
    "get_semantic_model_bim",
    "get_object_level_security",
    #'list_annotations',
    #'list_columns',
    "list_dashboards",
    "list_dataflow_storage_accounts",
    #'list_datamarts',
    #'list_datapipelines',
    #'list_eventstreams',
    #'list_kpis',
    #'list_kqldatabases',
    #'list_kqlquerysets',
    "list_lakehouses",
    #'list_mirroredwarehouses',
    #'list_mlexperiments',
    #'list_mlmodels',
    #'list_relationships',
    #'list_sqlendpoints',
    #'list_tables',
    "list_warehouses",
    "list_workspace_role_assignments",
    "create_warehouse",
    "update_item",
    "create_abfss_path",
    "format_dax_object_name",
    "create_relationship_name",
    "save_as_delta_table",
    "generate_embedded_filter",
    "get_direct_lake_sql_endpoint",
    "resolve_lakehouse_id",
    "resolve_lakehouse_name",
    "resolve_dataset_id",
    "resolve_dataset_name",
    "resolve_report_id",
    "resolve_report_name",
    #'language_validate',
    # "model_auto_build",
    "model_bpa_rules",
    "run_model_bpa",
    "measure_dependency_tree",
    "get_measure_dependencies",
    "get_model_calc_dependencies",
    "export_model_to_onelake",
    "qso_sync",
    "qso_sync_status",
    "set_qso",
    "list_qso_settings",
    "disable_qso",
    "set_semantic_model_storage_format",
    "set_workspace_default_storage_format",
    "refresh_semantic_model",
    "cancel_dataset_refresh",
    "translate_semantic_model",
    "vertipaq_analyzer",
    #'visualize_vertipaq',
    "import_vertipaq_analyzer",
    "list_semantic_model_objects",
    "list_shortcuts",
    "list_custom_pools",
    "create_custom_pool",
    "update_custom_pool",
    "assign_workspace_to_capacity",
    "unassign_workspace_from_capacity",
    "get_spark_settings",
    "update_spark_settings",
    "add_user_to_workspace",
    "delete_user_from_workspace",
    "update_workspace_user",
    "list_workspace_users",
    "assign_workspace_to_dataflow_storage",
    "list_capacities",
]
